"""Pre-defined pytest fixtures to be used in tests of the dependent apps.

This plugin is loaded automatically when this module is installed which makes the fixtures
available without further configuration.
"""

import contextlib
from unittest import mock

import pytest

# Check whether flask_jwt_extended (optional dependency) is installed.
JWT_INSTALLED = False
try:
    from . import jwt  # pylint: disable=unused-import

    JWT_INSTALLED = True
except ModuleNotFoundError:
    pass


if JWT_INSTALLED:
    import flask

    def _push_session_to_context(session, context):
        # pylint: disable=protected-access
        context._jwt_extended_jwt = session.to_token_data()
        context._jwt_extended_jwt_user = {"loaded_user": session}
        return None, context

    @pytest.fixture(name="jwt_inject_session")
    def fixture_inject_session():
        """Define a helper function to mock out JWT authentication checks."""

        @contextlib.contextmanager
        def inject_session(session):
            with mock.patch("flask_jwt_extended.verify_jwt_in_request") as mock_verify:
                mock_verify.side_effect = lambda *args, **kwargs: _push_session_to_context(
                    session, flask.g
                )
                yield

        return inject_session



"""Useful wrappers for requests."""

# Also provide exceptions to make them available with a single import.
from requests import exceptions  # pylint: disable=unused-import

from . import auth, sessions

Session = sessions.Session
__all__ = ["auth", "exceptions", "Session"]



"""Tests for the JWT auth-app utilities."""

import flask
import flask_jwt_extended
import pytest

from .. import jwt


def test_ancestors():
    """Run a few scenarios for the ancestors calculation."""
    assert jwt.ancestors(["root>parent>org1"]) == {"root", "root>parent"}
    assert jwt.ancestors(["root>parent>org1", "root>parent>org2"]) == {"root", "root>parent"}
    assert jwt.ancestors(["root>parent>org1", "root>other>org"]) == {
        "root",
        "root>parent",
        "root>other",
    }
    assert jwt.ancestors(["root>parent", "root>parent>org"]) == {"root", "root>parent"}

class SessionTest:
    """Unit-tests for the Session class."""

    def test_init_generates_session_id(self):
        """Test that creating a new session automatically generates a unique session id."""
        session1 = jwt.Session("user-id", {"org": []})
        session2 = jwt.Session("user-id", {"org": []})
        assert session1.session_id
        assert session2.session_id
        assert session1.session_id != session2.session_id

    def test_custom_uuid_init(self):
        """Test passing a custom session uuid in the constructor."""
        uuid = "24861915-4617-4dc9-ac0e-2326c7538abc"
        session = jwt.Session("user-id", {"org": []}, uuid)
        data = session.to_token_data()
        assert data["user_claims"]["session_id"] == uuid

    def test_create_session_from_token(self):
        """Test creating a session from a token."""
        token = {
            "identity": "7fd8ecf2-c1fa-43fa-8661-3eafaec457b0",
            "user_claims": {
                "roles": {"org1": ["admin"]},
                "session_id": "24861915-4617-4dc9-ac0e-2326c7538abc",
            },
        }
        session = jwt.Session.from_raw_token(token)
        assert session.to_token_data() == token

    def test_has_any_role_of_inherits_permissions(self):
        """Test permission inheritance."""
        roles = {
            "root": ["root-role"],
            "root>parent": ["parent-role"],
            "root>parent>org": ["org-role"],
            "root>other-parent": ["other-role"],
            "other-root": ["other-role"],
        }
        session = jwt.Session("user-id", roles)
        assert session.has_any_role_of(["root-role"], "root>parent>org")
        assert session.has_any_role_of(["parent-role"], "root>parent>org")
        assert session.has_any_role_of(["org-role"], "root>parent>org")
        assert not session.has_any_role_of(["other-role"], "root>parent>org")
        assert session.has_any_role_of([], "root")
        assert session.has_any_role_of([], "root>parent")
        assert session.has_any_role_of([], "root>parent>org")
        assert not session.has_any_role_of([], "unknown")

    def test_organizations_for_roles(self):
        """Test that it always returns a minimal set."""
        roles = {
            "root": ["root-role"],
            "root>parent": ["parent-role"],
            "root>parent>org": ["org-role"],
            "root>other-parent": ["other-role"],
            "other-root": ["other-role"],
        }
        session = jwt.Session("user-id", roles)
        assert session.organizations_for_roles(["root-role", "other-role"]) == {
            "root",
            "other-root",
        }
        assert session.organizations_for_roles(["parent-role", "org-role"]) == {
            "root>parent",
        }


@pytest.fixture(name="protected_app")
def fixture_protected_app():
    """Create an app with one JWT protected route."""
    app = flask.Flask("test")
    app.config["JWT_SECRET_KEY"] = "test-secret-key"
    jwt.manager.init_app(app)

    @app.route("/", methods=["GET", "OPTIONS"])
    @jwt.required(admitted_roles=["app"])
    def dump_session():
        session = jwt.get_current_session()
        return flask.jsonify(session.to_token_data())

    @app.route("/optional")
    @jwt.required(optional=True)
    def dump_session2():
        session = jwt.get_current_session()
        return flask.jsonify(session.to_token_data())

    return app


def test_getting_session_data_in_authorized_request(protected_app):
    """Send an authorized request to the endpoint."""
    session = jwt.Session("user-id", {"organization": ["app"]})
    with protected_app.app_context():
        token = flask_jwt_extended.create_access_token(session)

    with protected_app.test_client() as client:
        response = client.get("/", headers={"Authorization": "Bearer " + token})
    assert response.status_code == 200
    assert response.json["identity"] == "user-id"
    assert response.json["user_claims"]["roles"] == {"organization": ["app"]}
    assert response.json["user_claims"]["session_id"]


def test_getting_injected_session_data(protected_app, jwt_inject_session):
    """Inject session using the settings."""
    session = jwt.Session("user-id", {"organization": ["app"]})

    with jwt_inject_session(session):
        with protected_app.test_client() as client:
            response = client.get("/")

    assert response.status_code == 200
    assert response.json["identity"] == "user-id"
    assert response.json["user_claims"]["roles"] == {"organization": ["app"]}


def test_denying_access_without_an_admitted_role(protected_app, jwt_inject_session):
    """Test the error response given when the JWT doesn’t have any of the admitted roles."""
    session = jwt.Session("user-id", {"organization": "not-app"})

    with jwt_inject_session(session):
        with protected_app.test_client() as client:
            response = client.get("/")

    assert response.status_code == 403
    assert "The user has none of the admitted roles" in response.data.decode()


def test_exempt_request_method_for_non_optional_wrapper(protected_app):
    """Test that OPTION requests get an anonymous session on protected routes."""
    with protected_app.test_client() as client:
        response = client.options("/")

    assert response.status_code == 403


def test_anonymous_session_with_optional(protected_app):
    """Test that a session is created on-the-fly for anonymous users."""
    with protected_app.test_client() as client:
        response = client.get("/optional")
    assert response.status_code == 200
    assert response.json["identity"] is None
    assert response.json["user_claims"]["roles"] == {}


def test_anonymous_session_with_organization(protected_app):
    """Test that sessions for anonymous requests read the x-ist-org header."""
    with protected_app.test_client() as client:
        # Showcase that the header name is not case-sensitive.
        response = client.get("/optional", headers={"X-IsT-OrG": "test-org"})
    assert response.status_code == 200
    assert response.json["identity"] is None
    assert response.json["user_claims"]["roles"] == {"test-org": []}



"""Wrapper for the requests Session object."""

import requests


class Session(requests.Session):
    """Wrap the `requests.Session` to add default timeouts."""

    timeout = None

    def request(self, *args, **kwargs):
        """Send a request with a a default timeout for the session."""
        kwargs.setdefault("timeout", self.timeout)
        return super().request(*args, **kwargs)



"""Basic Flask App with a few enhancements."""

import os

from flask import Flask
from werkzeug.utils import import_string

from moflask.logging import init_logger


class BaseApp(Flask):
    """Flask App."""

    @staticmethod
    def config_obj_from_env(env=None):
        """Find a config object that matches the current environment."""
        env = env or os.getenv("FLASK_ENV", "production").capitalize()
        path = os.getenv("FLASK_CONFIG_OBJECT", "settings." + env + "Config")
        return import_string(path)()

    def __init__(self, *args, **kwargs):
        """Create a new app object."""
        config = kwargs.pop("config", None)
        env = kwargs.pop("env", None)
        super().__init__(*args, **kwargs)
        # load app config
        # 1. from an import path specfied in the FLASK_CONFIG_OBJECT
        #    environmentfrom variable. Default: settings.DevelopmentConfig
        # 2. A config object passed as keyword argument: config=Object
        self.config.from_object(self.config_obj_from_env(env))
        self.config.from_object(config)
        init_logger(self)

        if not self.sanity_check():
            raise RuntimeError("Sanity checks failed. Aborting!")

        self.init_sentry()

    def init_sentry(self):
        """Initialize Sentry."""
        if self.config.get("SENTRY_DSN", False):
            from raven.contrib.flask import Sentry

            self.sentry = Sentry(self)

    @staticmethod
    def sanity_check():
        """Check sanity."""
        return True



"""JWT session handling.

This module provides the common code-base needed for consuming auth-app JWTs.

The JWT’s claims provide a user identifier, an organization and the current roles of the user.
"""

import functools
import itertools
import uuid
from typing import Iterable, Set

import flask
import flask_jwt_extended
import werkzeug.exceptions

get_current_session = flask_jwt_extended.get_current_user


def iterate_parents(org: str, include_self=False):
    """Iterate over all parent organizations."""
    parts = org.split(">")
    if not include_self:
        parts.pop()
    return itertools.accumulate(parts, lambda a, b: f"{a}>{b}")


def ancestors(orgs: Iterable[str]) -> Set[str]:
    """Get all the ancestors of all the given organizations.

    Returns:
        A set of all organizations which are ancestors of at least one of the passed organizations.
    """
    return frozenset(itertools.chain.from_iterable(iterate_parents(org) for org in orgs))


class Session:
    """Object that represents a user’s session for a specific organization."""

    @classmethod
    def from_raw_token(cls, token: dict):
        """Create a session from raw JWT token data."""
        claims = token["user_claims"]
        return cls(token["identity"], claims["roles"], claims["session_id"])

    @classmethod
    def create_anonymous_session(cls):
        """Create an anonymous session from the request headers."""
        org = flask.request.headers.get("x-ist-org", None)
        return cls(None, {org: []} if org else {})

    def __init__(
        self, identity: str, roles: dict[str, Iterable[str]] = None, custom_uuid: str = None
    ):
        """Create a new session instance."""
        self.identity = identity
        self.roles = {org: frozenset(r) for org, r in roles.items()} if roles else {}
        self.session_id = custom_uuid or str(uuid.uuid4())

    def has_any_role_of(self, roles: Iterable[str], org: str = None) -> bool:
        """Check if the session has any of the given roles for the organization.

        Args:
            roles: The admitted roles. If empty then any list of roles will pass the check.
            org: Check roles for this organization or its parents. If None is passed the method
            returns True if the session has any of the roles for any organization.
        """
        roles = frozenset(roles)
        orgs = iterate_parents(org, include_self=True) if org is not None else self.roles.keys()
        return any(not roles or self.roles[o] & roles for o in orgs if o in self.roles)

    def organizations_for_roles(self, admitted_roles: Iterable[str]) -> Set[str]:
        """Get organizations for which the session has at least one of the given roles.

        This method is usually used to determine which ressources the session should have
        access to.

        Note that passing an empty list of admitted_roles means that any list of roles for an
        organization satisfies the condition, even an empty one.

        Returns:
            A set of organizations for which the session has any of the given roles. If both a
            parent and a parent>sub organization would be part of the set, then only the parent is
            returned.
            This is the minimal set of organizations so that has_any_role_of() would return True
            for all of them and any of their sub-organizations.
        """
        roles = frozenset(admitted_roles)
        orgs = frozenset(o for o, r in self.roles.items() if not roles or r & roles)
        cleaned = frozenset(o for o in orgs if not any(p in orgs for p in iterate_parents(o)))
        return cleaned

    def to_token_data(self):
        """Return the session object turned into the token data structure.

        This is the inverse of from_raw_token().
        """
        return {
            "identity": self.identity,
            "user_claims": {
                "session_id": self.session_id,
                "roles": {org: list(roles) for org, roles in self.roles.items()},
            },
        }


class Manager(flask_jwt_extended.JWTManager):
    """A JWTManager subclass with a bit of custom behaviour."""

    @staticmethod
    def _push_context_callback(session: Session, context):  # pylint: disable=method-hidden
        """Implement a default push context callback doing nothing."""

    def push_context(self, session):
        """Push data to the request context."""
        self._push_context_callback(session, flask.g)

    def push_context_callback(self, callback):
        """Register a callback for pushing data to the request context.

        Most commonly this callback would load a user or organization object from the database to be
        used later in the request. The callback can throw werkzeug.exceptions to interrupt the
        request handling.
        """
        self._push_context_callback = callback


manager = Manager()


@manager.user_lookup_loader
def create_session(_jwt_header, jwt_data):
    """Create a session object from token data.

    The session is made available in either of:
    - flask_jwt_extended.current_user
    - flask_jwt_extended.get_current_user()
    - moflask.jwt.get_current_session()
    with the latter being considered the most idiomatic.
    """
    return Session.from_raw_token(jwt_data)


@manager.additional_claims_loader
def session_to_token(session: Session):
    """Generate JWT claims that represent the session."""
    return session.to_token_data()


# Deactivate the default callback. session_to_token takes care of everything.
manager.user_identity_loader(lambda session: None)


def check_roles(session: Session, admitted_roles: Iterable[str], org: str = None):
    """Check whether the session has any of the admitted roles."""
    if not session.has_any_role_of(admitted_roles, org):
        raise werkzeug.exceptions.Forbidden("The user has none of the admitted roles.")


def required(admitted_roles: Iterable[str] = None, optional=False, **kwargs):
    """Require JWT authentication for a route.

    This decorator is a replacement for all of the flask_jwt_extended decorators.

    If you decorate an endpoint with this, it will ensure that the requester
    has a valid access token before allowing the endpoint to be called.

    The behaviour can be configured using the keyword parameters:
    - admitted_roles: Only allow session with any of these roles to access the route.
    """
    decorator_kwargs = kwargs

    def wrap(wrapped_fn):
        @functools.wraps(wrapped_fn)
        def wrapper(*args, **kwargs):
            if flask_jwt_extended.verify_jwt_in_request(optional=optional, **decorator_kwargs):
                session = get_current_session()
            else:  # Authentication is optional or the request method is exempt.
                session = Session.create_anonymous_session()
                # pylint: disable=protected-access
                flask.g._jwt_extended_jwt_user = {"loaded_user": session}
            manager.push_context(session)
            if session and admitted_roles is not None:
                check_roles(session, admitted_roles)
            return wrapped_fn(*args, **kwargs)

        return wrapper

    return wrap



"""Auth middlewares for the requests library."""

import flask
import requests

from . import rest


class AuthAppClient(rest.Client):
    """A REST client for retrieving JWTs from the auth-app."""

    AUTH_API_VERSION = "v1"

    @classmethod
    def from_app(cls, app=None):
        """Create a new instance by reading the config from the Flask app config."""
        app = app or flask.current_app
        return cls(
            app.config["IMPACT_STACK_API_URL"] + "/auth/" + cls.AUTH_API_VERSION,
            app.config["IMPACT_STACK_API_KEY"],
        )

    def __init__(self, base_url, api_key):
        """Create a new client instance."""
        super().__init__(base_url)
        self.api_key = api_key

    def get_token(self):
        """Use the API token to get a new JWT."""
        response = self.post("token", json=self.api_key)
        return response.json()["token"]


class AuthAppMiddleware:
    """Middleware for authenticating using JWT tokens.

    The middleware transparently requests an API-token from the auth-app on-demand.
    """

    # pylint: disable=too-few-public-methods
    # For now the middleware is very simple, but in the future it should also take care of caching
    # the JWT and renew if needed.

    def __init__(self, client=None):
        """Create new auth-app requests auth middleware."""
        self.client = client or AuthAppClient.from_app()

    def __call__(self, request: requests.PreparedRequest):
        """Add the JWT token to the request."""
        request.headers["Authorization"] = "Bearer " + self.client.get_token()
        return request
